"use strict";

exports.__esModule = true;
exports["default"] = {
  //- builtin-prepass
  "minification.constantFolding": require("babel-plugin-constant-folding"),

  //- builtin-pre
  strict: require("./other/strict"),
  eval: require("babel-plugin-eval"),
  _hoistDirectives: require("./internal/hoist-directives"),
  "minification.removeDebugger": require("babel-plugin-remove-debugger"),
  "minification.removeConsole": require("babel-plugin-remove-console"),
  "utility.inlineEnvironmentVariables": require("babel-plugin-inline-environment-variables"),
  "minification.deadCodeElimination": require("babel-plugin-dead-code-elimination"),
  _modules: require("./internal/modules"),
  "react.displayName": require("babel-plugin-react-display-name"),
  "es6.spec.modules": require("./es6/spec.modules"),
  "es6.spec.arrowFunctions": require("./es6/spec.arrow-functions"),
  "es6.spec.templateLiterals": require("./es6/spec.template-literals"),
  "es6.templateLiterals": require("babel-plugin-es2015-template-literals"),
  "es6.literals": require("babel-plugin-es2015-literals"),
  "validation.undeclaredVariableCheck": require("babel-plugin-undeclared-variables-check"),

  //- builtin-basic
  // this is where the bulk of the ES6 transformations take place, none of them require traversal state
  // so they can all be concatenated together for performance
  "spec.functionName": require("./spec/function-name"),
  "es7.classProperties": require("babel-plugin-class-properties"),
  "es7.trailingFunctionCommas": require("babel-plugin-trailing-function-commas"),
  "es7.asyncFunctions": require("babel-plugin-async-functions"),
  "es7.decorators": require("babel-plugin-decorators"),
  "es6.arrowFunctions": require("babel-plugin-es2015-arrow-functions"),
  "spec.blockScopedFunctions": require("babel-plugin-es2015-block-scoped-functions"),
  "optimisation.react.constantElements": require("babel-plugin-react-constant-elements"),
  "optimisation.react.inlineElements": require("babel-plugin-react-inline-elements"),
  "es7.comprehensions": require("babel-plugin-comprehensions"),
  "es6.classes": require("babel-plugin-es2015-classes"),
  asyncToGenerator: require("babel-plugin-async-to-generator"),
  bluebirdCoroutines: require("babel-plugin-bluebird-coroutines"),
  "es6.objectSuper": require("babel-plugin-es2015-bject-super"),
  "es7.objectRestSpread": require("babel-plugin-object-rest-spread"),
  "es7.exponentiationOperator": require("babel-plugin-exponentiation-operator"),
  "es5.properties.mutators": require("babel-plugin-property-mutators"),
  "es6.properties.shorthand": require("babel-plugin-es2015-shorthand-properties"),
  "es6.properties.computed": require("babel-plugin-es2015-computed-properties"),
  "optimisation.flow.forOf": require("./optimisation/flow.for-of"),
  "es6.forOf": require("babel-plugin-es2015-for-of"),
  "es6.regex.sticky": require("babel-plugin-es2015-sticky-regex"),
  "es6.regex.unicode": require("babel-plugin-es2015-unicode-regex"),
  "es6.constants": require("babel-plugin-es2015-constants"),
  "es7.exportExtensions": require("babel-plugin-export-extensions"),
  "spec.protoToAssign": require("babel-plugin-proto-to-assign"),
  "es7.doExpressions": require("babel-plugin-do-expressions"),
  "es6.spec.symbols": require("./es6/spec.symbols"),
  "es7.functionBind": require("babel-plugin-function-bind"),
  "spec.undefinedToVoid": require("babel-plugin-undefined-to-void"),

  //- builtin-advanced
  "es6.spread": require("babel-plugin-es2015-spread"),
  "es6.parameters": require("babel-plugin-es2015-parameters"),
  "es6.destructuring": require("babel-plugin-es2015-destructuring"),
  "es6.blockScoping": require("babel-plugin-es2015-block-scoping"),
  reactCompat: require("babel-plugin-react-jsx-compat"),
  react: require("babel-plugin-react-jsx"),
  regenerator: require("./other/regenerator"),

  // es6 syntax transformation is **forbidden** past this point since regenerator will chuck a massive
  // hissy fit

  //- builtin-modules
  runtime: require("babel-plugin-runtime"),
  "es6.modules": require("./es6/modules"),
  _moduleFormatter: require("./internal/module-formatter"),

  //- builtin-trailing
  // these clean up the output and do finishing up transformations, it's important to note that by this
  // stage you can't import any new modules or insert new ES6 as all those transformers have already
  // been ran
  "es6.tailCall": require("babel-plugin-es2015-tail-call"),
  _shadowFunctions: require("./internal/shadow-functions"),
  "es3.propertyLiterals": require("babel-plugin-es3-property-literals"),
  "es3.memberExpressionLiterals": require("babel-plugin-es3-member-expression-literals"),
  "minification.memberExpressionLiterals": require("babel-plugin-member-expression-literals"),
  "minification.propertyLiterals": require("babel-plugin-property-literals"),
  _blockHoist: require("./internal/block-hoist"),
  jscript: require("babel-plugin-jscript"),
  flow: require("babel-plugin-flow-strip-types"),
  "optimisation.modules.system": require("./optimisation/modules.system")
};
module.exports = exports["default"];